# Cơ Chế Cánh Tay Cơ Giới Trong Game Tự Động Hóa


Trong quá trình phát triển tựa game tự động hóa có tên chưa công bố, nhóm chúng tôi đã xây dựng hệ thống vận chuyển tương đồng với Factorio ở nhiều khía cạnh. Tuy không sử dụng băng chuyền (conveyor belt) như giải pháp chính, nhưng hệ thống cánh tay cơ giới (mechanical arm) đã được triển khai để xử lý các tác vụ di chuyển vật phẩm trong phạm vi ngắn.

**Nguyên lý hoạt động cơ bản**  
Cánh tay cơ giới thực hiện nhiệm vụ chuyển dời vật phẩm giữa hai điểm cố định theo một hướng xác định trước. Thành phần quan trọng nhất là các hộp chứa (container) và chính cánh tay, được tách biệt thành hệ thống độc lập để tối ưu hóa hiệu năng và dễ bảo trì. Việc thiết kế này cho phép hệ thống vận hành song song với các hệ thống sản xuất hay logistics khác, chỉ cập nhật kết quả cuối cùng cho các hệ thống còn lại.

**Kiến trúc hệ thống**  
Dù sử dụng mô hình lập trình dựa trên thực thể/thành phần/hệ thống (ECS), chúng tôi chọn cách triển khai riêng cho hệ thống cánh tay. Mỗi hộp chứa được đánh dấu bằng ID duy nhất, đồng thời liên kết với các Entity trong ECS để đảm bảo khả năng truy cập nhanh. Các loại hộp chứa bao gồm:
- Hộp chứa vật phẩm thông thường
- Khu vực đầu vào/đầu ra của máy chế tạo
- Ngăn chứa nhiên liệu của buồng đốt

Hệ thống cánh tay được cập nhật độc lập thông qua một chu kỳ tick riêng biệt. Sau khi xử lý hoàn tất, các hệ thống khác sẽ nhận biết thay đổi trong hộp chứa mà không cần theo dõi quá trình vận chuyển chi tiết.

**Độ phức tạp ẩn giấu trong cơ chế vận chuyển**  
Dù nguyên lý đơn giản "lấy từ A - đặt tại B", nhưng thực tế cần xử lý nhiều tình huống phức tạp:
1. **Lọc vật phẩm**: Mỗi cánh tay có bộ quy tắc trắng/đen xác định vật phẩm được phép vận chuyển. Ví dụ, cánh tay chuyên dụng chỉ vận chuyển quặng hoặc nhiên liệu.
2. **Hạn chế theo loại**: Các máy chế tạo chỉ chấp nhận vật liệu đầu vào cụ thể, buộc cánh tay phải tính toán lựa chọn vật phẩm phù hợp.
3. **Tương tác đa thùng chứa**: Một thực thể (Entity) có thể chứa nhiều hộp chứa vật phẩm. Ví dụ: máy chế tạo kết hợp buồng đốt và khu vực chế tạo riêng biệt.

**Giai đoạn vận hành của cánh tay cơ giới**  
1. **Chờ đợi (0-1 tick)**: Khi không có vật phẩm phù hợp hoặc điểm đến đã đầy.
2. **Lấy vật (1 tick)**: Tức thời hoàn tất trong một chu kỳ tick.
3. **Di chuyển (2-5 ticks)**: Tùy theo khoảng cách và tốc độ quay.
4. **Đặt vật (1 tick)**: Thành công nếu không gian mục tiêu còn chỗ.

**Thiết kế cấu trúc dữ liệu tối ưu**  
Để tối ưu hóa hiệu năng, chúng tôi thiết kế hệ thống "khe chứa" (slot) với cấu trúc 16 byte như sau:

```cpp
struct Slot {
    ID         : dword  // Định danh duy nhất
    NextID     : dword  // Liên kết trong danh sách vòng
    Bộ lọc     : word   // Mã số bộ lọc vật phẩm
    Loại vật   : word   // Mã loại vật phẩm
    Số lượng   : word   // Số lượng hiện tại
    Sức chứa   : word   // Tối đa có thể chứa
};
```

**Các ưu điểm nổi bật**:
- **Bộ nhớ liền mạch**: Các khe liên tiếp trong cùng hộp chứa được lưu trữ gần nhau để tối ưu cache.
- **Quản lý hiệu quả**: Sử dụng danh sách rác (freelist) tái sử dụng khe trống.
- **Tính linh hoạt**: Hệ thống lọc vật phẩm cho phép thiết lập quy tắc phức tạp qua các lớp bộ lọc.

**Cơ chế tìm kiếm vật phẩm thông minh**  
Cánh tay cơ giới chỉ lưu trữ ID của hai khe mục tiêu (vào/ra) thay vì toàn bộ hộp chứa. Khi gặp tình huống khe hiện tại không thể lấy/đặt vật phẩm, hệ thống sẽ duyệt qua danh sách vòng bằng con trỏ NextID để tìm khe phù hợp tiếp theo. Điều này giúp giảm thiểu thao tác truy cập bộ nhớ và tăng tốc độ xử lý.

**Xử lý ngoại lệ thông minh**  
Trong trường hợp đặc biệt như:
- Sức chứa thay đổi theo công thức chế tạo (ví dụ: máy chế tạo cấp cao có giới hạn chồng chất riêng)
- Bộ lọc động dựa trên trạng thái hệ thống
Hệ thống được thiết kế để dễ dàng mở rộng thông qua các lớp bộ lọc tùy chỉnh, bảo đảm tính tách biệt giữa các hệ thống con.

Giải pháp này không chỉ tối ưu hiệu năng mà còn tạo nền tảng mở cho các tính năng mở rộng trong tương lai như:
- Cánh tay đa nhiệm xử lý nhiều vật phẩm
- Hệ thống ưu tiên vận chuyển thông minh
- Giao diện trực quan theo dõi lộ trình vật phẩm

Việc tách biệt logic vận chuyển khỏi hệ thống chính giúp giảm độ phức tạp, đồng thời mở ra khả năng tối ưu hóa sâu hơn cho các tình huống đặc thù trong game.

---

> Tác giả:   
> URL: http://apkpm.com/posts/12/factorio_inserter/  

